# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from typing import TYPE_CHECKING, List
from ..protobufs import main_pb2
from ..units import Units
from ..call import call, call_async
from .oscilloscope_data import OscilloscopeData

if TYPE_CHECKING:
    from .device import Device


class Oscilloscope:
    """
    Provides a convenient way to control the oscilloscope data recording feature of some devices.
    The oscilloscope can record the values of some settings over time at high resolution.
    """

    @property
    def device(self) -> 'Device':
        """
        Device that this Oscilloscope measures.
        """
        return self._device

    def __init__(self, device: 'Device'):
        self._device = device

    def add_channel(
            self,
            axis: int,
            setting: str
    ) -> None:
        """
        Select a setting to be recorded.

        Args:
            axis: The 1-based index of the axis to record the value from.
            setting: The name of a setting to record.
        """
        request = main_pb2.OscilloscopeAddChannelRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = axis
        request.setting = setting
        call("oscilloscope/add_channel", request)

    async def add_channel_async(
            self,
            axis: int,
            setting: str
    ) -> None:
        """
        Select a setting to be recorded.

        Args:
            axis: The 1-based index of the axis to record the value from.
            setting: The name of a setting to record.
        """
        request = main_pb2.OscilloscopeAddChannelRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = axis
        request.setting = setting
        await call_async("oscilloscope/add_channel", request)

    def clear(
            self
    ) -> None:
        """
        Clear the list of channels to record.
        """
        request = main_pb2.DeviceEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        call("oscilloscope/clear_channels", request)

    async def clear_async(
            self
    ) -> None:
        """
        Clear the list of channels to record.
        """
        request = main_pb2.DeviceEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        await call_async("oscilloscope/clear_channels", request)

    def get_timebase(
            self,
            unit: Units = Units.NATIVE
    ) -> float:
        """
        Get the current sampling interval.

        Args:
            unit: Unit of measure to represent the timebase in.

        Returns:
            The current sampling interval in the selected time units.
        """
        request = main_pb2.DeviceGetSettingRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.setting = "scope.timebase"
        request.unit = unit.value
        response = main_pb2.DoubleResponse()
        call("device/get_setting", request, response)
        return response.value

    async def get_timebase_async(
            self,
            unit: Units = Units.NATIVE
    ) -> float:
        """
        Get the current sampling interval.

        Args:
            unit: Unit of measure to represent the timebase in.

        Returns:
            The current sampling interval in the selected time units.
        """
        request = main_pb2.DeviceGetSettingRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.setting = "scope.timebase"
        request.unit = unit.value
        response = main_pb2.DoubleResponse()
        await call_async("device/get_setting", request, response)
        return response.value

    def set_timebase(
            self,
            interval: float,
            unit: Units = Units.NATIVE
    ) -> None:
        """
        Set the sampling interval.

        Args:
            interval: Sample interval for the next oscilloscope recording. Minimum value is 100µs.
            unit: Unit of measure the timebase is represented in.
        """
        request = main_pb2.DeviceSetSettingRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.setting = "scope.timebase"
        request.value = interval
        request.unit = unit.value
        call("device/set_setting", request)

    async def set_timebase_async(
            self,
            interval: float,
            unit: Units = Units.NATIVE
    ) -> None:
        """
        Set the sampling interval.

        Args:
            interval: Sample interval for the next oscilloscope recording. Minimum value is 100µs.
            unit: Unit of measure the timebase is represented in.
        """
        request = main_pb2.DeviceSetSettingRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.setting = "scope.timebase"
        request.value = interval
        request.unit = unit.value
        await call_async("device/set_setting", request)

    def get_delay(
            self,
            unit: Units = Units.NATIVE
    ) -> float:
        """
        Get the delay before oscilloscope recording starts.

        Args:
            unit: Unit of measure to represent the delay in.

        Returns:
            The current start delay in the selected time units.
        """
        request = main_pb2.DeviceGetSettingRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.setting = "scope.delay"
        request.unit = unit.value
        response = main_pb2.DoubleResponse()
        call("device/get_setting", request, response)
        return response.value

    async def get_delay_async(
            self,
            unit: Units = Units.NATIVE
    ) -> float:
        """
        Get the delay before oscilloscope recording starts.

        Args:
            unit: Unit of measure to represent the delay in.

        Returns:
            The current start delay in the selected time units.
        """
        request = main_pb2.DeviceGetSettingRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.setting = "scope.delay"
        request.unit = unit.value
        response = main_pb2.DoubleResponse()
        await call_async("device/get_setting", request, response)
        return response.value

    def set_delay(
            self,
            interval: float,
            unit: Units = Units.NATIVE
    ) -> None:
        """
        Set the sampling start delay.

        Args:
            interval: Delay time between triggering a recording and the first data point being recorded.
            unit: Unit of measure the delay is represented in.
        """
        request = main_pb2.DeviceSetSettingRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.setting = "scope.delay"
        request.value = interval
        request.unit = unit.value
        call("device/set_setting", request)

    async def set_delay_async(
            self,
            interval: float,
            unit: Units = Units.NATIVE
    ) -> None:
        """
        Set the sampling start delay.

        Args:
            interval: Delay time between triggering a recording and the first data point being recorded.
            unit: Unit of measure the delay is represented in.
        """
        request = main_pb2.DeviceSetSettingRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.setting = "scope.delay"
        request.value = interval
        request.unit = unit.value
        await call_async("device/set_setting", request)

    def start(
            self
    ) -> None:
        """
        Trigger data recording.
        """
        request = main_pb2.OscilloscopeStartStopRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.start = True
        call("oscilloscope/start_or_stop", request)

    async def start_async(
            self
    ) -> None:
        """
        Trigger data recording.
        """
        request = main_pb2.OscilloscopeStartStopRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.start = True
        await call_async("oscilloscope/start_or_stop", request)

    def stop(
            self
    ) -> None:
        """
        End data recording if currently in progress.
        """
        request = main_pb2.OscilloscopeStartStopRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.start = False
        call("oscilloscope/start_or_stop", request)

    async def stop_async(
            self
    ) -> None:
        """
        End data recording if currently in progress.
        """
        request = main_pb2.OscilloscopeStartStopRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.start = False
        await call_async("oscilloscope/start_or_stop", request)

    def read(
            self
    ) -> List[OscilloscopeData]:
        """
        Reads the last-recorded data from the oscilloscope. Will block until any in-progress recording completes.

        Returns:
            Array of recorded channel data arrays, in the order added.
        """
        request = main_pb2.DeviceEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        response = main_pb2.OscilloscopeReadResponse()
        call("oscilloscope/read", request, response)
        return list(map(OscilloscopeData, response.data_ids))

    async def read_async(
            self
    ) -> List[OscilloscopeData]:
        """
        Reads the last-recorded data from the oscilloscope. Will block until any in-progress recording completes.

        Returns:
            Array of recorded channel data arrays, in the order added.
        """
        request = main_pb2.DeviceEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        response = main_pb2.OscilloscopeReadResponse()
        await call_async("oscilloscope/read", request, response)
        return list(map(OscilloscopeData, response.data_ids))



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from enum import Enum


class CommandCode(Enum):
    """
    Named constants for all Zaber Binary protocol commands.
    """

    RESET = 0
    HOME = 1
    RENUMBER = 2
    STORE_CURRENT_POSITION = 16
    RETURN_STORED_POSITION = 17
    MOVE_TO_STORED_POSITION = 18
    MOVE_ABSOLUTE = 20
    MOVE_RELATIVE = 21
    MOVE_AT_CONSTANT_SPEED = 22
    STOP = 23
    SET_ACTIVE_AXIS = 25
    SET_AXIS_DEVICE_NUMBER = 26
    SET_AXIS_INVERSION = 27
    SET_AXIS_VELOCITY_PROFILE = 28
    SET_AXIS_VELOCITY_SCALE = 29
    LOAD_EVENT_INSTRUCTION = 30
    RETURN_EVENT_INSTRUCTION = 31
    SET_CALIBRATION_MODE = 33
    SET_JOYSTICK_CALIBRATION_MODE = 33
    READ_OR_WRITE_MEMORY = 35
    RESTORE_SETTINGS = 36
    SET_MICROSTEP_RESOLUTION = 37
    SET_RUNNING_CURRENT = 38
    SET_HOLD_CURRENT = 39
    SET_DEVICE_MODE = 40
    SET_HOME_SPEED = 41
    SET_START_SPEED = 41
    SET_TARGET_SPEED = 42
    SET_ACCELERATION = 43
    SET_MAXIMUM_POSITION = 44
    SET_CURRENT_POSITION = 45
    SET_MAXIMUM_RELATIVE_MOVE = 46
    SET_HOME_OFFSET = 47
    SET_ALIAS_NUMBER = 48
    SET_LOCK_STATE = 49
    RETURN_DEVICE_ID = 50
    RETURN_FIRMWARE_VERSION = 51
    RETURN_POWER_SUPPLY_VOLTAGE = 52
    RETURN_SETTING = 53
    RETURN_STATUS = 54
    ECHO_DATA = 55
    RETURN_FIRMWARE_BUILD = 56
    RETURN_CURRENT_POSITION = 60
    RETURN_SERIAL_NUMBER = 63
    SET_PARK_STATE = 65
    SET_PERIPHERAL_ID = 66
    RETURN_DIGITAL_INPUT_COUNT = 67
    READ_DIGITAL_INPUT = 68
    READ_ALL_DIGITAL_INPUTS = 69
    RETURN_DIGITAL_OUTPUT_COUNT = 70
    READ_DIGITAL_OUTPUT = 71
    READ_ALL_DIGITAL_OUTPUTS = 72
    WRITE_DIGITAL_OUTPUT = 73
    WRITE_ALL_DIGITAL_OUTPUTS = 74
    RETURN_ANALOG_INPUT_COUNT = 75
    READ_ANALOG_INPUT = 76
    RETURN_ANALOG_OUTPUT_COUNT = 77
    MOVE_INDEX = 78
    SET_INDEX_DISTANCE = 79
    SET_CYCLE_DISTANCE = 80
    SET_FILTER_HOLDER_ID = 81
    RETURN_ENCODER_COUNT = 82
    RETURN_CALIBRATED_ENCODER_COUNT = 83
    RETURN_CALIBRATION_TYPE = 84
    RETURN_CALIBRATION_ERROR = 85
    RETURN_PERIPHERAL_SERIAL_NUMBER = 86
    SET_PERIPHERAL_SERIAL_NUMBER = 86
    FORCE_ABSOLUTE = 87
    FORCE_OFF = 88
    RETURN_ENCODER_POSITION = 89
    RETURN_PERIPHERAL_ID_PENDING = 91
    RETURN_PERIPHERAL_SERIAL_PENDING = 92
    ACTIVATE = 93
    SET_AUTO_REPLY_DISABLED_MODE = 101
    SET_MESSAGE_ID_MODE = 102
    SET_HOME_STATUS = 103
    SET_HOME_SENSOR_TYPE = 104
    SET_AUTO_HOME_DISABLED_MODE = 105
    SET_MINIMUM_POSITION = 106
    SET_KNOB_DISABLED_MODE = 107
    SET_KNOB_DIRECTION = 108
    SET_KNOB_MOVEMENT_MODE = 109
    SET_KNOB_JOG_SIZE = 110
    SET_KNOB_VELOCITY_SCALE = 111
    SET_KNOB_VELOCITY_PROFILE = 112
    SET_ACCELERATION_ONLY = 113
    SET_DECELERATION_ONLY = 114
    SET_MOVE_TRACKING_MODE = 115
    SET_MANUAL_MOVE_TRACKING_DISABLED_MODE = 116
    SET_MOVE_TRACKING_PERIOD = 117
    SET_CLOSED_LOOP_MODE = 118
    SET_SLIP_TRACKING_PERIOD = 119
    SET_STALL_TIMEOUT = 120
    SET_DEVICE_DIRECTION = 121
    SET_BAUDRATE = 122
    SET_PROTOCOL = 123
    CONVERT_TO_ASCII = 124
    RETURN_SYSTEM_ERROR_PRESENT = 126



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import TYPE_CHECKING, List
from ..call import call, call_async, call_sync

from ..protobufs import main_pb2
from .device_settings import DeviceSettings
from .axis import Axis
from .all_axes import AllAxes
from .warnings import Warnings
from .device_identity import DeviceIdentity
from .device_io import DeviceIO
from .response import Response
from .lockstep import Lockstep
from .oscilloscope import Oscilloscope
from .storage import DeviceStorage
from .stream import Stream
from .stream_buffer import StreamBuffer
from ..firmware_version import FirmwareVersion
from ..measurement import Measurement
from ..can_set_state_device_response import CanSetStateDeviceResponse

if TYPE_CHECKING:
    from .connection import Connection


class Device:
    """
    Represents the controller part of one device - may be either a standalone controller or an integrated controller.
    """

    @property
    def connection(self) -> 'Connection':
        """
        Connection of this device.
        """
        return self._connection

    @property
    def device_address(self) -> int:
        """
        The device address uniquely identifies the device on the connection.
        It can be configured or automatically assigned by the renumber command.
        """
        return self._device_address

    @property
    def settings(self) -> DeviceSettings:
        """
        Settings and properties of this device.
        """
        return self._settings

    @property
    def storage(self) -> DeviceStorage:
        """
        Key-value storage of this device.
        """
        return self._storage

    @property
    def io(self) -> DeviceIO:
        """
        I/O channels of this device.
        """
        return self._io

    @property
    def all_axes(self) -> AllAxes:
        """
        Virtual axis which allows you to target all axes of this device.
        """
        return self._all_axes

    @property
    def warnings(self) -> Warnings:
        """
        Warnings and faults of this device and all its axes.
        """
        return self._warnings

    @property
    def identity(self) -> DeviceIdentity:
        """
        Identity of the device.
        """
        return self.__retrieve_identity()

    @property
    def is_identified(self) -> bool:
        """
        Indicates whether or not the device has been identified.
        """
        return self.__retrieve_is_identified()

    @property
    def oscilloscope(self) -> Oscilloscope:
        """
        Oscilloscope recording helper for this device.
        """
        return self._oscilloscope

    @property
    def device_id(self) -> int:
        """
        Unique ID of the device hardware.
        """
        return self.identity.device_id

    @property
    def serial_number(self) -> int:
        """
        Serial number of the device.
        """
        return self.identity.serial_number

    @property
    def name(self) -> str:
        """
        Name of the product.
        """
        return self.identity.name

    @property
    def axis_count(self) -> int:
        """
        Number of axes this device has.
        """
        return self.identity.axis_count

    @property
    def firmware_version(self) -> FirmwareVersion:
        """
        Version of the firmware.
        """
        return self.identity.firmware_version

    @property
    def is_integrated(self) -> bool:
        """
        The device is an integrated product.
        """
        return self.identity.is_integrated

    def __init__(self, connection: 'Connection', device_address: int):
        self._connection = connection
        self._device_address = device_address
        self._settings = DeviceSettings(self)
        self._storage = DeviceStorage(self)
        self._io = DeviceIO(self)
        self._all_axes = AllAxes(self)
        self._warnings = Warnings(self, 0)
        self._oscilloscope = Oscilloscope(self)

    def identify(
            self
    ) -> DeviceIdentity:
        """
        Queries the device and the database, gathering information about the product.
        Without this information features such as unit conversions will not work.
        Usually, called automatically by detect devices method.

        Returns:
            Device identification data.
        """
        request = main_pb2.DeviceEmptyRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        response = main_pb2.DeviceIdentity()
        call("device/identify", request, response)
        return DeviceIdentity.from_protobuf(response)

    async def identify_async(
            self
    ) -> DeviceIdentity:
        """
        Queries the device and the database, gathering information about the product.
        Without this information features such as unit conversions will not work.
        Usually, called automatically by detect devices method.

        Returns:
            Device identification data.
        """
        request = main_pb2.DeviceEmptyRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        response = main_pb2.DeviceIdentity()
        await call_async("device/identify", request, response)
        return DeviceIdentity.from_protobuf(response)

    def generic_command(
            self,
            command: str,
            axis: int = 0,
            check_errors: bool = True,
            timeout: int = 0
    ) -> Response:
        """
        Sends a generic ASCII command to this device.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            axis: Optional axis number to send the command to.
            check_errors: Controls whether to throw an exception when the device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            A response to the command.
        """
        request = main_pb2.GenericCommandRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        request.command = command
        request.axis = axis
        request.check_errors = check_errors
        request.timeout = timeout
        response = main_pb2.GenericCommandResponse()
        call("interface/generic_command", request, response)
        return Response.from_protobuf(response)

    async def generic_command_async(
            self,
            command: str,
            axis: int = 0,
            check_errors: bool = True,
            timeout: int = 0
    ) -> Response:
        """
        Sends a generic ASCII command to this device.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            axis: Optional axis number to send the command to.
            check_errors: Controls whether to throw an exception when the device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            A response to the command.
        """
        request = main_pb2.GenericCommandRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        request.command = command
        request.axis = axis
        request.check_errors = check_errors
        request.timeout = timeout
        response = main_pb2.GenericCommandResponse()
        await call_async("interface/generic_command", request, response)
        return Response.from_protobuf(response)

    def generic_command_multi_response(
            self,
            command: str,
            axis: int = 0,
            check_errors: bool = True,
            timeout: int = 0
    ) -> List[Response]:
        """
        Sends a generic ASCII command to this device and expect multiple responses.
        Responses are returned in order of arrival.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            axis: Optional axis number to send the command to.
            check_errors: Controls whether to throw an exception when a device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            All responses to the command.
        """
        request = main_pb2.GenericCommandRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        request.command = command
        request.axis = axis
        request.check_errors = check_errors
        request.timeout = timeout
        response = main_pb2.GenericCommandResponseCollection()
        call("interface/generic_command_multi_response", request, response)
        return [Response.from_protobuf(a) for a in response.responses]

    async def generic_command_multi_response_async(
            self,
            command: str,
            axis: int = 0,
            check_errors: bool = True,
            timeout: int = 0
    ) -> List[Response]:
        """
        Sends a generic ASCII command to this device and expect multiple responses.
        Responses are returned in order of arrival.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            axis: Optional axis number to send the command to.
            check_errors: Controls whether to throw an exception when a device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            All responses to the command.
        """
        request = main_pb2.GenericCommandRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        request.command = command
        request.axis = axis
        request.check_errors = check_errors
        request.timeout = timeout
        response = main_pb2.GenericCommandResponseCollection()
        await call_async("interface/generic_command_multi_response", request, response)
        return [Response.from_protobuf(a) for a in response.responses]

    def generic_command_no_response(
            self,
            command: str,
            axis: int = 0
    ) -> None:
        """
        Sends a generic ASCII command to this device without expecting a response and without adding a message ID
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            axis: Optional axis number to send the command to.
                Specifying -1 omits the number completely.
        """
        request = main_pb2.GenericCommandRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        request.command = command
        request.axis = axis
        call("interface/generic_command_no_response", request)

    async def generic_command_no_response_async(
            self,
            command: str,
            axis: int = 0
    ) -> None:
        """
        Sends a generic ASCII command to this device without expecting a response and without adding a message ID
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            axis: Optional axis number to send the command to.
                Specifying -1 omits the number completely.
        """
        request = main_pb2.GenericCommandRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        request.command = command
        request.axis = axis
        await call_async("interface/generic_command_no_response", request)

    def get_axis(
            self,
            axis_number: int
    ) -> Axis:
        """
        Gets an Axis class instance which allows you to control a particular axis on this device.
        Axes are numbered from 1.

        Args:
            axis_number: Number of axis intended to control.

        Returns:
            Axis instance.
        """
        if axis_number <= 0:
            raise ValueError('Invalid value; physical axes are numbered from 1.')

        return Axis(self, axis_number)

    def get_lockstep(
            self,
            lockstep_group_id: int
    ) -> Lockstep:
        """
        Gets a Lockstep class instance which allows you to control a particular lockstep group on the device.

        Args:
            lockstep_group_id: The ID of the lockstep group to control. Lockstep group IDs start at one.

        Returns:
            Lockstep instance.
        """
        if lockstep_group_id <= 0:
            raise ValueError('Invalid value; lockstep groups are numbered from 1.')

        return Lockstep(self, lockstep_group_id)

    def prepare_command(
            self,
            command_template: str,
            *parameters: Measurement
    ) -> str:
        """
        Formats parameters into a command and performs unit conversions.
        Parameters in the command template are denoted by a question mark.
        Command returned is only valid for this device.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command_template: Template of a command to prepare. Parameters are denoted by question marks.
            parameters: Variable number of command parameters.

        Returns:
            Command with converted parameters.
        """
        request = main_pb2.PrepareCommandRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        request.command_template = command_template
        request.parameters.extend([Measurement.to_protobuf(a) for a in parameters])
        response = main_pb2.StringResponse()
        call_sync("device/prepare_command", request, response)
        return response.value

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the device.

        Returns:
            A string that represents the device.
        """
        request = main_pb2.DeviceEmptyRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        response = main_pb2.StringResponse()
        call_sync("device/device_to_string", request, response)
        return response.value

    def get_stream(
            self,
            stream_id: int
    ) -> Stream:
        """
        Gets a Stream class instance which allows you to control a particular stream on the device.

        Args:
            stream_id: The ID of the stream to control. Stream IDs start at one.

        Returns:
            Stream instance.
        """
        if stream_id <= 0:
            raise ValueError('Invalid value; streams are numbered from 1.')

        return Stream(self, stream_id)

    def get_stream_buffer(
            self,
            stream_buffer_id: int
    ) -> StreamBuffer:
        """
        Gets a StreamBuffer class instance which is a handle for a stream buffer on the device.

        Args:
            stream_buffer_id: The ID of the stream buffer to control. Stream buffer IDs start at one.

        Returns:
            StreamBuffer instance.
        """
        if stream_buffer_id <= 0:
            raise ValueError('Invalid value; stream buffers are numbered from 1.')

        return StreamBuffer(self, stream_buffer_id)

    def get_state(
            self
    ) -> str:
        """
        Returns a serialization of the current device state that can be saved and reapplied.

        Returns:
            A serialization of the current state of the device.
        """
        request = main_pb2.AxisEmptyRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        response = main_pb2.StringResponse()
        call("device/get_state", request, response)
        return response.value

    async def get_state_async(
            self
    ) -> str:
        """
        Returns a serialization of the current device state that can be saved and reapplied.

        Returns:
            A serialization of the current state of the device.
        """
        request = main_pb2.AxisEmptyRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        response = main_pb2.StringResponse()
        await call_async("device/get_state", request, response)
        return response.value

    def set_state(
            self,
            state: str,
            device_only: bool = False
    ) -> None:
        """
        Applies a saved state to an axis.

        Args:
            state: The state object to apply to this axis.
            device_only: If true, only device scope settings and features will be set.
        """
        request = main_pb2.SetStateRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        request.state = state
        request.device_only = device_only
        call("device/set_state", request)

    async def set_state_async(
            self,
            state: str,
            device_only: bool = False
    ) -> None:
        """
        Applies a saved state to an axis.

        Args:
            state: The state object to apply to this axis.
            device_only: If true, only device scope settings and features will be set.
        """
        request = main_pb2.SetStateRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        request.state = state
        request.device_only = device_only
        await call_async("device/set_state", request)

    def can_set_state(
            self,
            state: str
    ) -> CanSetStateDeviceResponse:
        """
        Checks if a state can be applied to a device and its peripherals.
        This only covers exceptions that can be determined statically such as mismatches of ID or version,
        the process of applying the state can still fail when running.

        Args:
            state: The state object to check against.

        Returns:
            An object listing errors that come up when trying to set the state.
        """
        request = main_pb2.CanSetStateRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        request.state = state
        response = main_pb2.CanSetStateDeviceResponse()
        call("device/can_set_state", request, response)
        return CanSetStateDeviceResponse.from_protobuf(response)

    async def can_set_state_async(
            self,
            state: str
    ) -> CanSetStateDeviceResponse:
        """
        Checks if a state can be applied to a device and its peripherals.
        This only covers exceptions that can be determined statically such as mismatches of ID or version,
        the process of applying the state can still fail when running.

        Args:
            state: The state object to check against.

        Returns:
            An object listing errors that come up when trying to set the state.
        """
        request = main_pb2.CanSetStateRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        request.state = state
        response = main_pb2.CanSetStateDeviceResponse()
        await call_async("device/can_set_state", request, response)
        return CanSetStateDeviceResponse.from_protobuf(response)

    def __retrieve_identity(
            self
    ) -> DeviceIdentity:
        """
        Returns identity.

        Returns:
            Device identity.
        """
        request = main_pb2.DeviceEmptyRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        response = main_pb2.DeviceIdentity()
        call_sync("device/get_identity", request, response)
        return DeviceIdentity.from_protobuf(response)

    def __retrieve_is_identified(
            self
    ) -> bool:
        """
        Returns whether or not the device have been identified.

        Returns:
            True if the device has already been identified. False otherwise.
        """
        request = main_pb2.DeviceEmptyRequest()
        request.interface_id = self.connection.interface_id
        request.device = self.device_address
        response = main_pb2.BoolResponse()
        call_sync("device/get_is_identified", request, response)
        return response.value



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from enum import Enum


class BinarySettings(Enum):
    """
    Named constants for all Zaber Binary protocol settings.
    """

    ACCELERATION = 0
    ACCELERATION_ONLY = 1
    ACTIVE_AXIS = 2
    ALIAS_NUMBER = 3
    ANALOG_INPUT_COUNT = 4
    ANALOG_OUTPUT_COUNT = 5
    AUTO_HOME_DISABLED_MODE = 6
    AUTO_REPLY_DISABLED_MODE = 7
    AXIS_DEVICE_NUMBER = 8
    AXIS_INVERSION = 9
    AXIS_VELOCITY_PROFILE = 10
    AXIS_VELOCITY_SCALE = 11
    BAUD_RATE = 12
    CALIBRATED_ENCODER_COUNT = 13
    CALIBRATION_ERROR = 14
    CALIBRATION_TYPE = 15
    CLOSED_LOOP_MODE = 16
    CURRENT_POSITION = 17
    CYCLE_DISTANCE = 18
    DECELERATION_ONLY = 19
    DEVICE_DIRECTION = 20
    DEVICE_ID = 21
    DEVICE_MODE = 22
    DIGITAL_INPUT_COUNT = 23
    DIGITAL_OUTPUT_COUNT = 24
    ENCODER_COUNT = 25
    ENCODER_POSITION = 26
    FILTER_HOLDER_ID = 27
    FIRMWARE_BUILD = 28
    FIRMWARE_VERSION = 29
    HOLD_CURRENT = 30
    HOME_OFFSET = 31
    HOME_SENSOR_TYPE = 32
    HOME_SPEED = 33
    HOME_STATUS = 34
    INDEX_DISTANCE = 35
    JOYSTICK_CALIBRATION_MODE = 36
    KNOB_DIRECTION = 37
    KNOB_DISABLED_MODE = 38
    KNOB_JOG_SIZE = 39
    KNOB_MOVEMENT_MODE = 40
    KNOB_VELOCITY_PROFILE = 41
    KNOB_VELOCITY_SCALE = 42
    LOCK_STATE = 43
    MANUAL_MOVE_TRACKING_DISABLED_MODE = 44
    MAXIMUM_POSITION = 45
    MAXIMUM_RELATIVE_MOVE = 46
    MESSAGE_ID_MODE = 47
    MICROSTEP_RESOLUTION = 48
    MINIMUM_POSITION = 49
    MOVE_TRACKING_MODE = 50
    MOVE_TRACKING_PERIOD = 51
    PARK_STATE = 52
    PERIPHERAL_ID = 53
    PERIPHERAL_ID_PENDING = 54
    PERIPHERAL_SERIAL_NUMBER = 55
    PERIPHERAL_SERIAL_PENDING = 56
    POWER_SUPPLY_VOLTAGE = 57
    PROTOCOL = 58
    RUNNING_CURRENT = 59
    SERIAL_NUMBER = 60
    SLIP_TRACKING_PERIOD = 61
    STALL_TIMEOUT = 62
    STATUS = 63
    SYSTEM_ERROR_PRESENT = 64
    TARGET_SPEED = 65
